from typing import Any

import pytest
from google.protobuf.struct_pb2 import Struct, Value

from lekko_client.evaluation.rules import evaluate_rule
from lekko_client.gen.lekko.rules.v1beta3.rules_pb2 import (
    Atom,
    CallExpression,
    ComparisonOperator,
    LogicalExpression,
    LogicalOperator,
    Rule,
)
from lekko_client.helpers import convert_context


@pytest.mark.parametrize(
    "context_val,namespace,config_name,expected",
    [
        (1, "ns_1", "feature_1", False),
        (2, "ns_1", "feature_1", False),
        (3, "ns_1", "feature_1", True),
        (4, "ns_1", "feature_1", False),
        (5, "ns_1", "feature_1", True),
        (101, "ns_1", "feature_1", True),
        (102, "ns_1", "feature_1", True),
        (103, "ns_1", "feature_1", False),
        (104, "ns_1", "feature_1", False),
        (105, "ns_1", "feature_1", True),
        (1, "ns_2", "feature_2", False),
        (2, "ns_2", "feature_2", True),
        (3, "ns_2", "feature_2", False),
        (4, "ns_2", "feature_2", False),
        (5, "ns_2", "feature_2", True),
        (101, "ns_2", "feature_2", True),
        (102, "ns_2", "feature_2", True),
        (103, "ns_2", "feature_2", False),
        (104, "ns_2", "feature_2", True),
        (105, "ns_2", "feature_2", True),
        (3.1415, "ns_1", "feature_1", False),
        (2.7182, "ns_1", "feature_1", False),
        (1.6180, "ns_1", "feature_1", True),
        (6.6261, "ns_1", "feature_1", True),
        (6.0221, "ns_1", "feature_1", False),
        (2.9979, "ns_1", "feature_1", True),
        (6.6730, "ns_1", "feature_1", False),
        (1.3807, "ns_1", "feature_1", True),
        (1.4142, "ns_1", "feature_1", True),
        (2.0000, "ns_1", "feature_1", False),
        (3.1415, "ns_2", "feature_2", True),
        (2.7182, "ns_2", "feature_2", False),
        (1.6180, "ns_2", "feature_2", True),
        (6.6261, "ns_2", "feature_2", False),
        (6.0221, "ns_2", "feature_2", False),
        (2.9979, "ns_2", "feature_2", False),
        (6.6730, "ns_2", "feature_2", False),
        (1.3807, "ns_2", "feature_2", False),
        (1.4142, "ns_2", "feature_2", True),
        (2.0000, "ns_2", "feature_2", False),
        ("hello", "ns_1", "feature_1", False),
        ("world", "ns_1", "feature_1", False),
        ("i", "ns_1", "feature_1", True),
        ("am", "ns_1", "feature_1", True),
        ("a", "ns_1", "feature_1", True),
        ("unit", "ns_1", "feature_1", False),
        ("test", "ns_1", "feature_1", True),
        ("case", "ns_1", "feature_1", True),
        ("for", "ns_1", "feature_1", False),
        ("bucket", "ns_1", "feature_1", False),
        ("hello", "ns_2", "feature_2", True),
        ("world", "ns_2", "feature_2", False),
        ("i", "ns_2", "feature_2", True),
        ("am", "ns_2", "feature_2", True),
        ("a", "ns_2", "feature_2", True),
        ("unit", "ns_2", "feature_2", False),
        ("test", "ns_2", "feature_2", True),
        ("case", "ns_2", "feature_2", False),
        ("for", "ns_2", "feature_2", False),
        ("bucket", "ns_2", "feature_2", False),
    ],
)
def test_bucket(context_val, namespace, config_name, expected):
    rule = Rule(call_expression=CallExpression(bucket=CallExpression.Bucket(context_key="key", threshold=50000)))
    context = convert_context({"key": context_val})
    assert evaluate_rule(rule, namespace, config_name, context) == expected


def test_bool_const():
    for b in [True, False]:
        rule = Rule(bool_const=b)
        assert evaluate_rule(rule, "ns1", "config1") == b


def test_present():
    rule = Rule(atom=Atom(context_key="age", comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_PRESENT))
    assert evaluate_rule(rule, "ns1", "config1") is False
    assert evaluate_rule(rule, "ns1", "config1", convert_context({"age": 10})) is True
    assert evaluate_rule(rule, "ns1", "config1", convert_context({"age": "not a number"})) is True


def convert_to_value(v: Any) -> Value:
    s = Struct()
    s.update({"key": v})
    return s.fields["key"]


@pytest.mark.parametrize(
    "test_atom,test_context,expected,raises",
    [
        (
            Atom(
                context_key="isprod",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                comparison_value=convert_to_value(True),
            ),
            convert_context({"isprod": True}),
            True,
            False,
        ),
        (
            Atom(
                context_key="isprod",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                comparison_value=convert_to_value(True),
            ),
            convert_context({"isprod": False}),
            False,
            False,
        ),
        (
            Atom(
                context_key="isprod",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                comparison_value=convert_to_value(True),
            ),
            convert_context({"isprod": "not a bool"}),
            False,
            True,
        ),
        (
            Atom(
                context_key="age",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                comparison_value=convert_to_value(12),
            ),
            convert_context({"age": 12}),
            True,
            False,
        ),
        (
            Atom(
                context_key="age",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                comparison_value=convert_to_value(12),
            ),
            convert_context({"age": 35}),
            False,
            False,
        ),
        (
            Atom(
                context_key="age",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                comparison_value=convert_to_value(12),
            ),
            convert_context({"age": 12.001}),
            False,
            False,
        ),
        (
            Atom(
                context_key="age",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                comparison_value=convert_to_value(12.001),
            ),
            convert_context({"age": 12.001}),
            True,
            False,
        ),
        (
            Atom(
                context_key="age",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                comparison_value=convert_to_value(12),
            ),
            convert_context({"age": "not a number"}),
            False,
            True,
        ),
        (
            Atom(
                context_key="age",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_NOT_EQUALS,
                comparison_value=convert_to_value(12),
            ),
            convert_context({"age": 25}),
            True,
            False,
        ),
        (
            Atom(
                context_key="age",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_NOT_EQUALS,
                comparison_value=convert_to_value(12),
            ),
            convert_context({"age": 12}),
            False,
            False,
        ),
        (
            Atom(
                context_key="age",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                comparison_value=convert_to_value(12),
            ),
            convert_context({}),  # not present
            False,
            False,
        ),
        (
            Atom(
                context_key="city",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                comparison_value=convert_to_value("Rome"),
            ),
            convert_context({"city": "Rome"}),
            True,
            False,
        ),
        (
            Atom(
                context_key="city",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                comparison_value=convert_to_value("Rome"),
            ),
            convert_context({"city": "rome"}),
            False,
            False,
        ),
        (
            Atom(
                context_key="city",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                comparison_value=convert_to_value("Rome"),
            ),
            convert_context({"city": "Paris"}),
            False,
            False,
        ),
        (
            Atom(
                context_key="city",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                comparison_value=convert_to_value("Rome"),
            ),
            convert_context({"city": 99}),
            False,
            True,
        ),
        (
            Atom(
                context_key="age",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_LESS_THAN,
                comparison_value=convert_to_value(12),
            ),
            convert_context({"age": 12}),
            False,
            False,
        ),
        (
            Atom(
                context_key="age",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_LESS_THAN,
                comparison_value=convert_to_value(12),
            ),
            convert_context({"age": 11}),
            True,
            False,
        ),
        (
            Atom(
                context_key="age",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_LESS_THAN_OR_EQUALS,
                comparison_value=convert_to_value(12),
            ),
            convert_context({"age": 12}),
            True,
            False,
        ),
        (
            Atom(
                context_key="age",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_GREATER_THAN_OR_EQUALS,
                comparison_value=convert_to_value(12),
            ),
            convert_context({"age": 12}),
            True,
            False,
        ),
        (
            Atom(
                context_key="age",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_GREATER_THAN,
                comparison_value=convert_to_value(12),
            ),
            convert_context({"age": 12}),
            False,
            False,
        ),
        (
            Atom(
                context_key="age",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_GREATER_THAN,
                comparison_value=convert_to_value(12),
            ),
            convert_context({"age": 13}),
            True,
            False,
        ),
        (
            Atom(
                context_key="age",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_GREATER_THAN,
                comparison_value=convert_to_value(12),
            ),
            convert_context({"age": "not a number"}),
            False,
            True,
        ),
        (
            Atom(
                context_key="city",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_CONTAINED_WITHIN,
                comparison_value=convert_to_value(["Rome", "Paris"]),
            ),
            convert_context({"city": "London"}),
            False,
            False,
        ),
        (
            Atom(
                context_key="city",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_CONTAINED_WITHIN,
                comparison_value=convert_to_value(["Rome", "Paris"]),
            ),
            convert_context({"city": "Rome"}),
            True,
            False,
        ),
        (
            Atom(
                context_key="city",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_CONTAINED_WITHIN,
                comparison_value=convert_to_value(["Rome", "Paris"]),
            ),
            convert_context({}),  # not present
            False,
            False,
        ),
        (
            Atom(
                context_key="city",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_CONTAINED_WITHIN,
                comparison_value=convert_to_value(["Rome", "Paris"]),
            ),
            convert_context({"city": "rome"}),
            False,
            False,
        ),
        (
            Atom(
                context_key="city",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_STARTS_WITH,
                comparison_value=convert_to_value("Ro"),
            ),
            convert_context({"city": "Rome"}),
            True,
            False,
        ),
        (
            Atom(
                context_key="city",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_STARTS_WITH,
                comparison_value=convert_to_value("Ro"),
            ),
            convert_context({"city": "London"}),
            False,
            False,
        ),
        (
            Atom(
                context_key="city",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_STARTS_WITH,
                comparison_value=convert_to_value("Ro"),
            ),
            convert_context({"city": "rome"}),
            False,
            False,
        ),
        (
            Atom(
                context_key="city",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_STARTS_WITH,
                comparison_value=convert_to_value("Ro"),
            ),
            None,  # not present
            False,
            False,
        ),
        (
            Atom(
                context_key="city",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_ENDS_WITH,
                comparison_value=convert_to_value("me"),
            ),
            convert_context({"city": "Rome"}),
            True,
            False,
        ),
        (
            Atom(
                context_key="city",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_ENDS_WITH,
                comparison_value=convert_to_value("me"),
            ),
            convert_context({"city": "London"}),
            False,
            False,
        ),
        (
            Atom(
                context_key="city",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_CONTAINS,
                comparison_value=convert_to_value(""),
            ),
            convert_context({"city": "Rome"}),
            True,
            False,
        ),
        (
            Atom(
                context_key="city",
                comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_CONTAINS,
                comparison_value=convert_to_value("foo"),
            ),
            convert_context({"city": "Rome"}),
            False,
            False,
        ),
    ],
)
def test_atom(test_atom, test_context, expected, raises):
    rule = Rule(atom=test_atom)
    if raises:
        with pytest.raises(Exception):
            evaluate_rule(rule, "ns1", "config1", test_context)
    else:
        assert evaluate_rule(rule, "ns1", "config1", test_context) == expected


@pytest.mark.parametrize(
    "test_atoms,logical_op,test_context,expected,raises",
    [
        (
            [
                Atom(
                    context_key="age",
                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_LESS_THAN,
                    comparison_value=convert_to_value(5),
                ),
                Atom(
                    context_key="age",
                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_GREATER_THAN,
                    comparison_value=convert_to_value(10),
                ),
            ],
            LogicalOperator.LOGICAL_OPERATOR_OR,
            convert_context({"age": 8}),
            False,
            False,
        ),
        (
            [
                Atom(
                    context_key="age",
                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_LESS_THAN,
                    comparison_value=convert_to_value(5),
                ),
                Atom(
                    context_key="age",
                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_GREATER_THAN,
                    comparison_value=convert_to_value(10),
                ),
            ],
            LogicalOperator.LOGICAL_OPERATOR_OR,
            convert_context({"age": 12}),
            True,
            False,
        ),
        (
            [
                Atom(
                    context_key="age",
                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_LESS_THAN,
                    comparison_value=convert_to_value(5),
                ),
                Atom(
                    context_key="city",
                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                    comparison_value=convert_to_value("Rome"),
                ),
            ],
            LogicalOperator.LOGICAL_OPERATOR_AND,
            convert_context({"age": 8, "city": "Rome"}),
            False,
            False,
        ),
        (
            [
                Atom(
                    context_key="age",
                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_LESS_THAN,
                    comparison_value=convert_to_value(5),
                ),
                Atom(
                    context_key="city",
                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                    comparison_value=convert_to_value("Rome"),
                ),
            ],
            LogicalOperator.LOGICAL_OPERATOR_AND,
            convert_context({"age": 3, "city": "Rome"}),
            True,
            False,
        ),
        (
            [
                Atom(
                    context_key="age",
                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_LESS_THAN,
                    comparison_value=convert_to_value(5),
                ),
                Atom(
                    context_key="city",
                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                    comparison_value=convert_to_value("Rome"),
                ),
            ],
            LogicalOperator.LOGICAL_OPERATOR_UNSPECIFIED,
            convert_context({"age": 3}),
            True,
            False,
        ),
        (
            [
                Atom(
                    context_key="age",
                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_LESS_THAN,
                    comparison_value=convert_to_value(5),
                ),
                Atom(
                    context_key="city",
                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                    comparison_value=convert_to_value("Rome"),
                ),
            ],
            LogicalOperator.LOGICAL_OPERATOR_AND,
            convert_context({"age": 3}),
            False,
            False,
        ),
        (
            [
                Atom(
                    context_key="age",
                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_LESS_THAN,
                    comparison_value=convert_to_value(5),
                )
            ],
            LogicalOperator.LOGICAL_OPERATOR_AND,
            convert_context({"age": 3}),
            True,
            False,
        ),
        (
            [
                Atom(
                    context_key="age",
                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_LESS_THAN,
                    comparison_value=convert_to_value(5),
                )
            ],
            LogicalOperator.LOGICAL_OPERATOR_OR,
            convert_context({"age": 3}),
            True,
            False,
        ),
        ([], LogicalOperator.LOGICAL_OPERATOR_AND, convert_context({}), False, True),
        (
            [
                Atom(
                    context_key="age",
                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_LESS_THAN,
                    comparison_value=convert_to_value(5),
                ),
                Atom(
                    context_key="city",
                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                    comparison_value=convert_to_value("Rome"),
                ),
                Atom(
                    context_key="age",
                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                    comparison_value=convert_to_value(8),
                ),
            ],
            LogicalOperator.LOGICAL_OPERATOR_AND,
            convert_context({"age": 8}),
            False,
            False,
        ),
        (
            [
                Atom(
                    context_key="age",
                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_LESS_THAN,
                    comparison_value=convert_to_value(5),
                ),
                Atom(
                    context_key="city",
                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                    comparison_value=convert_to_value("Rome"),
                ),
                Atom(
                    context_key="age",
                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                    comparison_value=convert_to_value(8),
                ),
            ],
            LogicalOperator.LOGICAL_OPERATOR_OR,
            convert_context({"age": 8}),
            True,
            False,
        ),
    ],
)
def test_logical_op(test_atoms, logical_op, test_context, expected, raises):
    rule = Rule(
        logical_expression=LogicalExpression(
            rules=[Rule(atom=atom) for atom in test_atoms], logical_operator=logical_op
        )
    )

    if raises:
        with pytest.raises(Exception):
            evaluate_rule(rule, "ns_1", "feature_1", test_context)
    else:
        assert evaluate_rule(rule, "ns_1", "feature_1", test_context) == expected



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2022 Lekko Technologies, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import lekko_client.gen.lekko.feature.v1beta1.feature_pb2
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class StaticFeature(google.protobuf.message.Message):
    """Represents a statically parsed feature.
    This model is meant to contain all sorts of data that is available
    to us via inspecting the starlark file, but may not be available
    post-compilation. E.g. what import statements were defined, and what
    comments were written surrounding certain expressions.
    The goal is to add to this model any information we wish to surface
    to the UI or that is needed for static mutation of a feature.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KEY_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    IMPORTS_FIELD_NUMBER: builtins.int
    FEATURE_FIELD_NUMBER: builtins.int
    FEATURE_OLD_FIELD_NUMBER: builtins.int
    key: builtins.str
    type: lekko_client.gen.lekko.feature.v1beta1.feature_pb2.FeatureType.ValueType
    @property
    def imports(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ImportStatement]: ...
    @property
    def feature(self) -> global___FeatureStruct:
        """Experimental protobuf message that is a representation
        of the statically parsed starlark AST.
        """
    @property
    def feature_old(self) -> lekko_client.gen.lekko.feature.v1beta1.feature_pb2.Feature:
        """For backwards compatibility. Eventually, the materialized
        model will be deprecated in favor of a statically parsed model
        """
    def __init__(
        self,
        *,
        key: builtins.str = ...,
        type: lekko_client.gen.lekko.feature.v1beta1.feature_pb2.FeatureType.ValueType = ...,
        imports: collections.abc.Iterable[global___ImportStatement] | None = ...,
        feature: global___FeatureStruct | None = ...,
        feature_old: lekko_client.gen.lekko.feature.v1beta1.feature_pb2.Feature | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["feature", b"feature", "feature_old", b"feature_old"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["feature", b"feature", "feature_old", b"feature_old", "imports", b"imports", "key", b"key", "type", b"type"]) -> None: ...

global___StaticFeature = StaticFeature

@typing_extensions.final
class FeatureStruct(google.protobuf.message.Message):
    """Represents everything stored in the feature struct in starlark.
    i.e. `feature(...)`
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    META_FIELD_NUMBER: builtins.int
    DESCRIPTION_FIELD_NUMBER: builtins.int
    DEFAULT_FIELD_NUMBER: builtins.int
    RULES_FIELD_NUMBER: builtins.int
    @property
    def meta(self) -> global___StarMeta: ...
    description: builtins.str
    @property
    def default(self) -> global___StarExpr: ...
    @property
    def rules(self) -> global___Rules: ...
    def __init__(
        self,
        *,
        meta: global___StarMeta | None = ...,
        description: builtins.str = ...,
        default: global___StarExpr | None = ...,
        rules: global___Rules | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["default", b"default", "meta", b"meta", "rules", b"rules"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["default", b"default", "description", b"description", "meta", b"meta", "rules", b"rules"]) -> None: ...

global___FeatureStruct = FeatureStruct

@typing_extensions.final
class Rules(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    META_FIELD_NUMBER: builtins.int
    RULES_FIELD_NUMBER: builtins.int
    @property
    def meta(self) -> global___StarMeta: ...
    @property
    def rules(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Rule]: ...
    def __init__(
        self,
        *,
        meta: global___StarMeta | None = ...,
        rules: collections.abc.Iterable[global___Rule] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["meta", b"meta"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["meta", b"meta", "rules", b"rules"]) -> None: ...

global___Rules = Rules

@typing_extensions.final
class Rule(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    META_FIELD_NUMBER: builtins.int
    CONDITION_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    @property
    def meta(self) -> global___StarMeta: ...
    condition: builtins.str
    @property
    def value(self) -> global___StarExpr: ...
    def __init__(
        self,
        *,
        meta: global___StarMeta | None = ...,
        condition: builtins.str = ...,
        value: global___StarExpr | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["meta", b"meta", "value", b"value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["condition", b"condition", "meta", b"meta", "value", b"value"]) -> None: ...

global___Rule = Rule

@typing_extensions.final
class ImportStatement(google.protobuf.message.Message):
    """An assignment expression in starlark, e.g. `x = 1`."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    META_FIELD_NUMBER: builtins.int
    LHS_FIELD_NUMBER: builtins.int
    OPERATOR_FIELD_NUMBER: builtins.int
    LINE_BREAK_FIELD_NUMBER: builtins.int
    RHS_FIELD_NUMBER: builtins.int
    @property
    def meta(self) -> global___StarMeta: ...
    @property
    def lhs(self) -> global___IdentExpr:
        """the variable that the import is assigned to"""
    operator: builtins.str
    """e.g. `=`"""
    line_break: builtins.bool
    """Line break between operator and RHS"""
    @property
    def rhs(self) -> global___ImportExpr: ...
    def __init__(
        self,
        *,
        meta: global___StarMeta | None = ...,
        lhs: global___IdentExpr | None = ...,
        operator: builtins.str = ...,
        line_break: builtins.bool = ...,
        rhs: global___ImportExpr | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["lhs", b"lhs", "meta", b"meta", "rhs", b"rhs"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["lhs", b"lhs", "line_break", b"line_break", "meta", b"meta", "operator", b"operator", "rhs", b"rhs"]) -> None: ...

global___ImportStatement = ImportStatement

@typing_extensions.final
class ImportExpr(google.protobuf.message.Message):
    """an import expression, e.g. `proto.package("google.protobuf")`"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    META_FIELD_NUMBER: builtins.int
    DOT_FIELD_NUMBER: builtins.int
    ARGS_FIELD_NUMBER: builtins.int
    @property
    def meta(self) -> global___StarMeta: ...
    @property
    def dot(self) -> global___DotExpr: ...
    @property
    def args(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        meta: global___StarMeta | None = ...,
        dot: global___DotExpr | None = ...,
        args: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["dot", b"dot", "meta", b"meta"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["args", b"args", "dot", b"dot", "meta", b"meta"]) -> None: ...

global___ImportExpr = ImportExpr

@typing_extensions.final
class DotExpr(google.protobuf.message.Message):
    """a dot expression, e.g. `proto.package`"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    META_FIELD_NUMBER: builtins.int
    X_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    @property
    def meta(self) -> global___StarMeta: ...
    x: builtins.str
    """The token that appears before the dot in the dot expression, e.g. `proto`"""
    name: builtins.str
    """The token that appears after the dot in the dot expression, e.g. `package`"""
    def __init__(
        self,
        *,
        meta: global___StarMeta | None = ...,
        x: builtins.str = ...,
        name: builtins.str = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["meta", b"meta"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["meta", b"meta", "name", b"name", "x", b"x"]) -> None: ...

global___DotExpr = DotExpr

@typing_extensions.final
class StarExpr(google.protobuf.message.Message):
    """A single starlark expression. May not be fully decomposed.
    Stores the string representation of the expression in the metadata.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    META_FIELD_NUMBER: builtins.int
    EXPRESSION_FIELD_NUMBER: builtins.int
    @property
    def meta(self) -> global___StarMeta: ...
    expression: builtins.str
    def __init__(
        self,
        *,
        meta: global___StarMeta | None = ...,
        expression: builtins.str = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["meta", b"meta"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["expression", b"expression", "meta", b"meta"]) -> None: ...

global___StarExpr = StarExpr

@typing_extensions.final
class IdentExpr(google.protobuf.message.Message):
    """A single token in starlark. E.g. True, False, or a variable name."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    META_FIELD_NUMBER: builtins.int
    TOKEN_FIELD_NUMBER: builtins.int
    @property
    def meta(self) -> global___StarMeta: ...
    token: builtins.str
    def __init__(
        self,
        *,
        meta: global___StarMeta | None = ...,
        token: builtins.str = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["meta", b"meta"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["meta", b"meta", "token", b"token"]) -> None: ...

global___IdentExpr = IdentExpr

@typing_extensions.final
class StarMeta(google.protobuf.message.Message):
    """Metadata commonly associated with any starlark expression"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    COMMENTS_FIELD_NUMBER: builtins.int
    MULTILINE_FIELD_NUMBER: builtins.int
    @property
    def comments(self) -> global___Comments: ...
    multiline: builtins.bool
    def __init__(
        self,
        *,
        comments: global___Comments | None = ...,
        multiline: builtins.bool = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["comments", b"comments"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["comments", b"comments", "multiline", b"multiline"]) -> None: ...

global___StarMeta = StarMeta

@typing_extensions.final
class Comments(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BEFORE_FIELD_NUMBER: builtins.int
    SUFFIX_FIELD_NUMBER: builtins.int
    AFTER_FIELD_NUMBER: builtins.int
    @property
    def before(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Comment]: ...
    @property
    def suffix(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Comment]: ...
    @property
    def after(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Comment]: ...
    def __init__(
        self,
        *,
        before: collections.abc.Iterable[global___Comment] | None = ...,
        suffix: collections.abc.Iterable[global___Comment] | None = ...,
        after: collections.abc.Iterable[global___Comment] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["after", b"after", "before", b"before", "suffix", b"suffix"]) -> None: ...

global___Comments = Comments

@typing_extensions.final
class Comment(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TOKEN_FIELD_NUMBER: builtins.int
    token: builtins.str
    def __init__(
        self,
        *,
        token: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["token", b"token"]) -> None: ...

global___Comment = Comment
